# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
import logging
from unittest import mock as async_mock

import pytest

from airflow.providers.google.cloud.triggers.cloud_sql import CloudSQLExportTrigger
from airflow.triggers.base import TriggerEvent

CLASSPATH = "airflow.providers.google.cloud.triggers.cloud_sql.CloudSQLExportTrigger"
TASK_ID = "test_task"
TEST_POLL_INTERVAL = 10
TEST_GCP_CONN_ID = "test-project"
HOOK_STR = "airflow.providers.google.cloud.hooks.cloud_sql.{}"
PROJECT_ID = "test_project_id"
OPERATION_NAME = "test_operation_name"
OPERATION_URL = (
    f"https://sqladmin.googleapis.com/sql/v1beta4/projects/{PROJECT_ID}/operations/{OPERATION_NAME}"
)


@pytest.fixture
def trigger():
    return CloudSQLExportTrigger(
        operation_name=OPERATION_NAME,
        project_id=PROJECT_ID,
        impersonation_chain=None,
        gcp_conn_id=TEST_GCP_CONN_ID,
        poke_interval=TEST_POLL_INTERVAL,
    )


class TestCloudSQLExportTrigger:
    def test_async_export_trigger_serialization_should_execute_successfully(self, trigger):
        """
        Asserts that the CloudSQLExportTrigger correctly serializes its arguments
        and classpath.
        """
        classpath, kwargs = trigger.serialize()
        assert classpath == CLASSPATH
        assert kwargs == {
            "operation_name": OPERATION_NAME,
            "project_id": PROJECT_ID,
            "impersonation_chain": None,
            "gcp_conn_id": TEST_GCP_CONN_ID,
            "poke_interval": TEST_POLL_INTERVAL,
        }

    @pytest.mark.asyncio
    @async_mock.patch(HOOK_STR.format("CloudSQLAsyncHook.get_operation"))
    async def test_async_export_trigger_on_success_should_execute_successfully(
        self, mock_get_operation, trigger
    ):
        """
        Tests the CloudSQLExportTrigger only fires once the job execution reaches a successful state.
        """
        mock_get_operation.return_value = {
            "status": "DONE",
            "name": OPERATION_NAME,
        }
        generator = trigger.run()
        actual = await generator.asend(None)
        assert (
            TriggerEvent(
                {
                    "operation_name": OPERATION_NAME,
                    "status": "success",
                }
            )
            == actual
        )

    @pytest.mark.asyncio
    @async_mock.patch(HOOK_STR.format("CloudSQLAsyncHook.get_operation"))
    async def test_async_export_trigger_running_should_execute_successfully(
        self, mock_get_operation, trigger, caplog
    ):
        """
        Test that CloudSQLExportTrigger does not fire while a job is still running.
        """

        mock_get_operation.return_value = {
            "status": "RUNNING",
            "name": OPERATION_NAME,
        }
        caplog.set_level(logging.INFO)
        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was not returned
        assert task.done() is False

        assert f"Operation status is RUNNING, sleeping for {TEST_POLL_INTERVAL} seconds." in caplog.text

        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @async_mock.patch(HOOK_STR.format("CloudSQLAsyncHook.get_operation"))
    async def test_async_export_trigger_error_should_execute_successfully(self, mock_get_operation, trigger):
        """
        Test that CloudSQLExportTrigger fires the correct event in case of an error.
        """
        mock_get_operation.return_value = {
            "status": "DONE",
            "name": OPERATION_NAME,
            "error": {"message": "test_error"},
        }

        expected_event = {
            "operation_name": OPERATION_NAME,
            "status": "error",
            "message": "test_error",
        }

        generator = trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent(expected_event) == actual

    @pytest.mark.asyncio
    @async_mock.patch(HOOK_STR.format("CloudSQLAsyncHook.get_operation"))
    async def test_async_export_trigger_exception_should_execute_successfully(
        self, mock_get_operation, trigger
    ):
        """
        Test that CloudSQLExportTrigger fires the correct event in case of an error.
        """
        mock_get_operation.side_effect = Exception("Test exception")

        generator = trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "failed", "message": "Test exception"}) == actual



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG for Google Cloud Storage ACL (Access Control List) operators.
"""

from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.gcs import (
    GCSBucketCreateAclEntryOperator,
    GCSCreateBucketOperator,
    GCSDeleteBucketOperator,
    GCSObjectCreateAclEntryOperator,
)
from airflow.providers.google.cloud.transfers.gcs_to_gcs import GCSToGCSOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
GCS_ACL_ENTITY = os.environ.get("SYSTEM_TESTS_GCS_ACL_ENTITY", "allUsers")

DAG_ID = "gcs_acl"

RESOURCES_BUCKET_NAME = "airflow-system-tests-resources"
BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}"
FILE_NAME = "example_upload.txt"
UPLOAD_FILE_PATH = f"gcs/{FILE_NAME}"

GCS_ACL_BUCKET_ROLE = "OWNER"
GCS_ACL_OBJECT_ROLE = "OWNER"


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["gcs", "acl", "example"],
) as dag:
    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket",
        bucket_name=BUCKET_NAME,
        project_id=PROJECT_ID,
        resource={"predefined_acl": "public_read_write"},
    )

    copy_file = GCSToGCSOperator(
        task_id="copy_example_gcs_file",
        source_bucket=RESOURCES_BUCKET_NAME,
        source_object=UPLOAD_FILE_PATH,
        destination_bucket=BUCKET_NAME,
        destination_object=FILE_NAME,
        exact_match=True,
    )

    # [START howto_operator_gcs_bucket_create_acl_entry_task]
    gcs_bucket_create_acl_entry_task = GCSBucketCreateAclEntryOperator(
        bucket=BUCKET_NAME,
        entity=GCS_ACL_ENTITY,
        role=GCS_ACL_BUCKET_ROLE,
        task_id="gcs_bucket_create_acl_entry_task",
    )
    # [END howto_operator_gcs_bucket_create_acl_entry_task]

    # [START howto_operator_gcs_object_create_acl_entry_task]
    gcs_object_create_acl_entry_task = GCSObjectCreateAclEntryOperator(
        bucket=BUCKET_NAME,
        object_name=FILE_NAME,
        entity=GCS_ACL_ENTITY,
        role=GCS_ACL_OBJECT_ROLE,
        task_id="gcs_object_create_acl_entry_task",
    )
    # [END howto_operator_gcs_object_create_acl_entry_task]

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket", bucket_name=BUCKET_NAME, trigger_rule=TriggerRule.ALL_DONE
    )

    (
        # TEST SETUP
        create_bucket
        >> copy_file
        # TEST BODY
        >> gcs_bucket_create_acl_entry_task
        >> gcs_object_create_acl_entry_task
        # TEST TEARDOWN,
        >> delete_bucket
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG for DataprocSubmitJobOperator with spark job
in deferrable mode.
"""

from __future__ import annotations

import os
from datetime import datetime

from google.api_core.retry import Retry

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.dataproc import (
    DataprocCreateClusterOperator,
    DataprocDeleteClusterOperator,
    DataprocSubmitJobOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
DAG_ID = "dataproc_spark_deferrable"
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

CLUSTER_NAME_BASE = f"cluster-{DAG_ID}".replace("_", "-")
CLUSTER_NAME_FULL = CLUSTER_NAME_BASE + f"-{ENV_ID}".replace("_", "-")
CLUSTER_NAME = CLUSTER_NAME_BASE if len(CLUSTER_NAME_FULL) >= 33 else CLUSTER_NAME_FULL

REGION = "europe-west1"

# Cluster definition
CLUSTER_CONFIG = {
    "master_config": {
        "num_instances": 1,
        "machine_type_uri": "n1-standard-4",
        "disk_config": {"boot_disk_type": "pd-standard", "boot_disk_size_gb": 32},
    },
    "worker_config": {
        "num_instances": 2,
        "machine_type_uri": "n1-standard-4",
        "disk_config": {"boot_disk_type": "pd-standard", "boot_disk_size_gb": 32},
    },
}

# Jobs definitions
# [START how_to_cloud_dataproc_spark_deferrable_config]
SPARK_JOB = {
    "reference": {"project_id": PROJECT_ID},
    "placement": {"cluster_name": CLUSTER_NAME},
    "spark_job": {
        "jar_file_uris": ["file:///usr/lib/spark/examples/jars/spark-examples.jar"],
        "main_class": "org.apache.spark.examples.SparkPi",
    },
}
# [END how_to_cloud_dataproc_spark_deferrable_config]


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "dataproc", "spark", "deferrable"],
) as dag:
    create_cluster = DataprocCreateClusterOperator(
        task_id="create_cluster",
        project_id=PROJECT_ID,
        cluster_config=CLUSTER_CONFIG,
        region=REGION,
        cluster_name=CLUSTER_NAME,
        retry=Retry(maximum=100.0, initial=10.0, multiplier=1.0),
        num_retries_if_resource_is_not_ready=3,
    )

    spark_task = DataprocSubmitJobOperator(
        task_id="spark_task", job=SPARK_JOB, region=REGION, project_id=PROJECT_ID, deferrable=True
    )

    delete_cluster = DataprocDeleteClusterOperator(
        task_id="delete_cluster",
        project_id=PROJECT_ID,
        cluster_name=CLUSTER_NAME,
        region=REGION,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    (
        # TEST SETUP
        create_cluster
        # TEST BODY
        >> spark_task
        # TEST TEARDOWN
        >> delete_cluster
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "teardown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from unittest import mock

from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.monitoring_v3 import AlertPolicy, NotificationChannel

from airflow.providers.google.cloud.operators.stackdriver import (
    StackdriverDeleteAlertOperator,
    StackdriverDeleteNotificationChannelOperator,
    StackdriverDisableAlertPoliciesOperator,
    StackdriverDisableNotificationChannelsOperator,
    StackdriverEnableAlertPoliciesOperator,
    StackdriverEnableNotificationChannelsOperator,
    StackdriverListAlertPoliciesOperator,
    StackdriverListNotificationChannelsOperator,
    StackdriverUpsertAlertOperator,
    StackdriverUpsertNotificationChannelOperator,
)

TEST_TASK_ID = "test-stackdriver-operator"
TEST_FILTER = "filter"
TEST_ALERT_POLICY_1 = {
    "combiner": "OR",
    "name": "projects/sd-project/alertPolicies/12345",
    "enabled": True,
    "display_name": "test display",
    "conditions": [
        {
            "condition_threshold": {
                "comparison": "COMPARISON_GT",
                "aggregations": [{"alignment_eriod": {"seconds": 60}, "per_series_aligner": "ALIGN_RATE"}],
            },
            "display_name": "Condition display",
            "name": "projects/sd-project/alertPolicies/123/conditions/456",
        }
    ],
}

TEST_ALERT_POLICY_2 = {
    "combiner": "OR",
    "name": "projects/sd-project/alertPolicies/6789",
    "enabled": False,
    "display_name": "test display",
    "conditions": [
        {
            "condition_threshold": {
                "comparison": "COMPARISON_GT",
                "aggregations": [{"alignment_period": {"seconds": 60}, "per_series_aligner": "ALIGN_RATE"}],
            },
            "display_name": "Condition display",
            "name": "projects/sd-project/alertPolicies/456/conditions/789",
        }
    ],
}

TEST_NOTIFICATION_CHANNEL_1 = {
    "displayName": "sd",
    "enabled": True,
    "labels": {"auth_token": "top-secret", "channel_name": "#channel"},
    "name": "projects/sd-project/notificationChannels/12345",
    "type": "slack",
}

TEST_NOTIFICATION_CHANNEL_2 = {
    "displayName": "sd",
    "enabled": False,
    "labels": {"auth_token": "top-secret", "channel_name": "#channel"},
    "name": "projects/sd-project/notificationChannels/6789",
    "type": "slack",
}


class TestStackdriverListAlertPoliciesOperator:
    @mock.patch("airflow.providers.google.cloud.operators.stackdriver.StackdriverHook")
    def test_execute(self, mock_hook):
        operator = StackdriverListAlertPoliciesOperator(task_id=TEST_TASK_ID, filter_=TEST_FILTER)
        mock_hook.return_value.list_alert_policies.return_value = [AlertPolicy(name="test-name")]
        result = operator.execute(context=mock.MagicMock())
        mock_hook.return_value.list_alert_policies.assert_called_once_with(
            project_id=None,
            filter_=TEST_FILTER,
            format_=None,
            order_by=None,
            page_size=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        assert result == [
            {
                "combiner": 0,
                "conditions": [],
                "display_name": "",
                "name": "test-name",
                "notification_channels": [],
                "severity": 0,
                "user_labels": {},
            }
        ]


class TestStackdriverEnableAlertPoliciesOperator:
    @mock.patch("airflow.providers.google.cloud.operators.stackdriver.StackdriverHook")
    def test_execute(self, mock_hook):
        operator = StackdriverEnableAlertPoliciesOperator(task_id=TEST_TASK_ID, filter_=TEST_FILTER)
        operator.execute(context=mock.MagicMock())
        mock_hook.return_value.enable_alert_policies.assert_called_once_with(
            project_id=None, filter_=TEST_FILTER, retry=DEFAULT, timeout=None, metadata=()
        )


class TestStackdriverDisableAlertPoliciesOperator:
    @mock.patch("airflow.providers.google.cloud.operators.stackdriver.StackdriverHook")
    def test_execute(self, mock_hook):
        operator = StackdriverDisableAlertPoliciesOperator(task_id=TEST_TASK_ID, filter_=TEST_FILTER)
        operator.execute(context=mock.MagicMock())
        mock_hook.return_value.disable_alert_policies.assert_called_once_with(
            project_id=None, filter_=TEST_FILTER, retry=DEFAULT, timeout=None, metadata=()
        )


class TestStackdriverUpsertAlertsOperator:
    @mock.patch("airflow.providers.google.cloud.operators.stackdriver.StackdriverHook")
    def test_execute(self, mock_hook):
        operator = StackdriverUpsertAlertOperator(
            task_id=TEST_TASK_ID, alerts=json.dumps({"policies": [TEST_ALERT_POLICY_1, TEST_ALERT_POLICY_2]})
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.return_value.upsert_alert.assert_called_once_with(
            alerts=json.dumps({"policies": [TEST_ALERT_POLICY_1, TEST_ALERT_POLICY_2]}),
            project_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestStackdriverDeleteAlertOperator:
    @mock.patch("airflow.providers.google.cloud.operators.stackdriver.StackdriverHook")
    def test_execute(self, mock_hook):
        operator = StackdriverDeleteAlertOperator(
            task_id=TEST_TASK_ID,
            name="test-alert",
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.return_value.delete_alert_policy.assert_called_once_with(
            name="test-alert", retry=DEFAULT, timeout=None, metadata=()
        )


class TestStackdriverListNotificationChannelsOperator:
    @mock.patch("airflow.providers.google.cloud.operators.stackdriver.StackdriverHook")
    def test_execute(self, mock_hook):
        operator = StackdriverListNotificationChannelsOperator(task_id=TEST_TASK_ID, filter_=TEST_FILTER)
        mock_hook.return_value.list_notification_channels.return_value = [
            NotificationChannel(name="test-123")
        ]

        result = operator.execute(context=mock.MagicMock())
        mock_hook.return_value.list_notification_channels.assert_called_once_with(
            project_id=None,
            filter_=TEST_FILTER,
            format_=None,
            order_by=None,
            page_size=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        # Depending on the version of google-apitools installed we might receive the response either with or
        # without mutation_records.
        assert result in [
            [
                {
                    "description": "",
                    "display_name": "",
                    "labels": {},
                    "name": "test-123",
                    "type_": "",
                    "user_labels": {},
                    "verification_status": 0,
                }
            ],
            [
                {
                    "description": "",
                    "display_name": "",
                    "labels": {},
                    "mutation_records": [],
                    "name": "test-123",
                    "type_": "",
                    "user_labels": {},
                    "verification_status": 0,
                }
            ],
        ]


class TestStackdriverEnableNotificationChannelsOperator:
    @mock.patch("airflow.providers.google.cloud.operators.stackdriver.StackdriverHook")
    def test_execute(self, mock_hook):
        operator = StackdriverEnableNotificationChannelsOperator(task_id=TEST_TASK_ID, filter_=TEST_FILTER)
        operator.execute(context=mock.MagicMock())
        mock_hook.return_value.enable_notification_channels.assert_called_once_with(
            project_id=None, filter_=TEST_FILTER, retry=DEFAULT, timeout=None, metadata=()
        )


class TestStackdriverDisableNotificationChannelsOperator:
    @mock.patch("airflow.providers.google.cloud.operators.stackdriver.StackdriverHook")
    def test_execute(self, mock_hook):
        operator = StackdriverDisableNotificationChannelsOperator(task_id=TEST_TASK_ID, filter_=TEST_FILTER)
        operator.execute(context=mock.MagicMock())
        mock_hook.return_value.disable_notification_channels.assert_called_once_with(
            project_id=None, filter_=TEST_FILTER, retry=DEFAULT, timeout=None, metadata=()
        )


class TestStackdriverUpsertChannelOperator:
    @mock.patch("airflow.providers.google.cloud.operators.stackdriver.StackdriverHook")
    def test_execute(self, mock_hook):
        operator = StackdriverUpsertNotificationChannelOperator(
            task_id=TEST_TASK_ID,
            channels=json.dumps({"channels": [TEST_NOTIFICATION_CHANNEL_1, TEST_NOTIFICATION_CHANNEL_2]}),
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.return_value.upsert_channel.assert_called_once_with(
            channels=json.dumps({"channels": [TEST_NOTIFICATION_CHANNEL_1, TEST_NOTIFICATION_CHANNEL_2]}),
            project_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestStackdriverDeleteNotificationChannelOperator:
    @mock.patch("airflow.providers.google.cloud.operators.stackdriver.StackdriverHook")
    def test_execute(self, mock_hook):
        operator = StackdriverDeleteNotificationChannelOperator(
            task_id=TEST_TASK_ID,
            name="test-channel",
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.return_value.delete_notification_channel.assert_called_once_with(
            name="test-channel", retry=DEFAULT, timeout=None, metadata=()
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Campaign Manager sensor."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.providers.common.compat.sdk import BaseSensorOperator
from airflow.providers.google.marketing_platform.hooks.campaign_manager import GoogleCampaignManagerHook

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class GoogleCampaignManagerReportSensor(BaseSensorOperator):
    """
    Check if report is ready.

    .. seealso::
        Check official API docs:
        https://developers.google.com/doubleclick-advertisers/rest/v4/reports/get

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleCampaignManagerReportSensor`

    :param profile_id: The DFA user profile ID.
    :param report_id: The ID of the report.
    :param file_id: The ID of the report file.
    :param api_version: The version of the api that will be requested, for example 'v4'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "profile_id",
        "report_id",
        "file_id",
        "impersonation_chain",
    )

    def poke(self, context: Context) -> bool:
        hook = GoogleCampaignManagerHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        response = hook.get_report(profile_id=self.profile_id, report_id=self.report_id, file_id=self.file_id)
        self.log.info("Report status: %s", response["status"])
        return response["status"] not in ("PROCESSING", "QUEUED")

    def __init__(
        self,
        *,
        profile_id: str,
        report_id: str,
        file_id: str,
        api_version: str = "v4",
        gcp_conn_id: str = "google_cloud_default",
        mode: str = "reschedule",
        poke_interval: int = 60 * 5,
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.mode = mode
        self.poke_interval = poke_interval
        self.profile_id = profile_id
        self.report_id = report_id
        self.file_id = file_id
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import logging
import os
import re
from contextlib import contextmanager
from io import StringIO
from tempfile import NamedTemporaryFile
from unittest import mock
from unittest.mock import ANY
from uuid import uuid4

import pytest
from google.auth.environment_vars import CREDENTIALS
from google.auth.exceptions import DefaultCredentialsError

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.utils.credentials_provider import (
    _DEFAULT_SCOPES,
    AIRFLOW_CONN_GOOGLE_CLOUD_DEFAULT,
    _get_project_id_from_service_account_email,
    _get_scopes,
    _get_target_principal_and_delegates,
    build_gcp_conn,
    get_credentials_and_project_id,
    provide_gcp_conn_and_credentials,
    provide_gcp_connection,
    provide_gcp_credentials,
)

ENV_VALUE = "test_env"
TEMP_VARIABLE = "temp_variable"
KEY = str(uuid4())
ENV_CRED = "temp_cred"
ACCOUNT_1_SAME_PROJECT = "account_1@project_id.iam.gserviceaccount.com"
ACCOUNT_2_SAME_PROJECT = "account_2@project_id.iam.gserviceaccount.com"
ACCOUNT_3_ANOTHER_PROJECT = "account_3@another_project_id.iam.gserviceaccount.com"
ANOTHER_PROJECT_ID = "another_project_id"
CRED_PROVIDER_LOGGER_NAME = "airflow.providers.google.cloud.utils.credentials_provider._CredentialProvider"
IDP_LINK = "http://example.com/idp"
CLIENT_ID = "your-client-id"
CLIENT_SECRET = "your-client-secret"
TEST_AUDIENCE = "test-audience"
TOKEN_TYPE = "urn:ietf:params:oauth:token-type:jwt"
ACCOUNT_IMPERSONATION = "http://example.com/impersonate"
CREDENTIAL_CONFIG_STRING_FILE = (
    '{"audience": "'
    + TEST_AUDIENCE
    + '", "subject_token_type": "'
    + TOKEN_TYPE
    + '", "service_account_impersonation_url": "'
    + ACCOUNT_IMPERSONATION
    + '"}'
)
CREDENTIAL_CONFIG_DICT_FILE = {
    "audience": TEST_AUDIENCE,
    "subject_token_type": TOKEN_TYPE,
    "service_account_impersonation_url": ACCOUNT_IMPERSONATION,
}


@pytest.fixture
def assert_no_logs(caplog):
    """
    Helper fixture for assert if any log message for the specific logger captured.

    This is workaround for fix issue with asyncio in ``test_disable_logging``, see:
        - https://github.com/apache/airflow/pull/26871
        - https://github.com/apache/airflow/pull/26973
        - https://github.com/apache/airflow/pull/36453
    """

    @contextmanager
    def wrapper(level: str, logger: str):
        with caplog.at_level(level=level, logger=logger):
            caplog.clear()
            yield
        if records := list(filter(lambda lr: lr[0] == logger, caplog.record_tuples)):
            msg = f"Did not expect any log message from logger={logger!r} but got:"
            for log_record in records:
                msg += f"\n * logger name: {log_record[0]!r}, level: {log_record[1]}, msg: {log_record[2]!r}"
            raise AssertionError(msg)

    return wrapper


class TestHelper:
    def test_build_gcp_conn_path(self):
        value = "test"
        conn = build_gcp_conn(key_file_path=value)
        assert conn == "google-cloud-platform://?key_path=test"

    def test_build_gcp_conn_scopes(self):
        value = ["test", "test2"]
        conn = build_gcp_conn(scopes=value)
        assert conn == "google-cloud-platform://?scope=test%2Ctest2"

    def test_build_gcp_conn_project(self):
        value = "test"
        conn = build_gcp_conn(project_id=value)
        assert conn == "google-cloud-platform://?projects=test"


class TestProvideGcpCredentials:
    @mock.patch.dict(os.environ, {CREDENTIALS: ENV_VALUE})
    @mock.patch("tempfile.NamedTemporaryFile")
    def test_provide_gcp_credentials_key_content(self, mock_file):
        file_dict = {"foo": "bar"}
        string_file = StringIO()
        file_content = json.dumps(file_dict)
        file_name = "/test/mock-file"
        mock_file_handler = mock_file.return_value.__enter__.return_value
        mock_file_handler.name = file_name
        mock_file_handler.write = string_file.write

        with provide_gcp_credentials(key_file_dict=file_dict):
            assert os.environ[CREDENTIALS] == file_name
            assert file_content == string_file.getvalue()
        assert os.environ[CREDENTIALS] == ENV_VALUE

    @mock.patch.dict(os.environ, {CREDENTIALS: ENV_VALUE})
    def test_provide_gcp_credentials_keep_environment(self):
        key_path = "/test/key-path"
        with provide_gcp_credentials(key_file_path=key_path):
            assert os.environ[CREDENTIALS] == key_path
        assert os.environ[CREDENTIALS] == ENV_VALUE


class TestProvideGcpConnection:
    @mock.patch.dict(os.environ, {AIRFLOW_CONN_GOOGLE_CLOUD_DEFAULT: ENV_VALUE})
    @mock.patch("airflow.providers.google.cloud.utils.credentials_provider.build_gcp_conn")
    def test_provide_gcp_connection(self, mock_builder):
        mock_builder.return_value = TEMP_VARIABLE
        path = "path/to/file.json"
        scopes = ["scopes"]
        project_id = "project_id"
        with provide_gcp_connection(path, scopes, project_id):
            mock_builder.assert_called_once_with(key_file_path=path, scopes=scopes, project_id=project_id)
            assert os.environ[AIRFLOW_CONN_GOOGLE_CLOUD_DEFAULT] == TEMP_VARIABLE
        assert os.environ[AIRFLOW_CONN_GOOGLE_CLOUD_DEFAULT] == ENV_VALUE


class TestProvideGcpConnAndCredentials:
    @mock.patch.dict(
        os.environ,
        {AIRFLOW_CONN_GOOGLE_CLOUD_DEFAULT: ENV_VALUE, CREDENTIALS: ENV_VALUE},
    )
    @mock.patch("airflow.providers.google.cloud.utils.credentials_provider.build_gcp_conn")
    def test_provide_gcp_conn_and_credentials(self, mock_builder):
        mock_builder.return_value = TEMP_VARIABLE
        path = "path/to/file.json"
        scopes = ["scopes"]
        project_id = "project_id"
        with provide_gcp_conn_and_credentials(path, scopes, project_id):
            mock_builder.assert_called_once_with(key_file_path=path, scopes=scopes, project_id=project_id)
            assert os.environ[AIRFLOW_CONN_GOOGLE_CLOUD_DEFAULT] == TEMP_VARIABLE
            assert os.environ[CREDENTIALS] == path
        assert os.environ[AIRFLOW_CONN_GOOGLE_CLOUD_DEFAULT] == ENV_VALUE
        assert os.environ[CREDENTIALS] == ENV_VALUE


class TestGetGcpCredentialsAndProjectId:
    test_scopes = _DEFAULT_SCOPES
    test_key_file = "KEY_PATH.json"
    test_project_id = "project_id"

    @mock.patch("google.auth.default", return_value=("CREDENTIALS", "PROJECT_ID"))
    def test_get_credentials_and_project_id_with_default_auth(self, mock_auth_default, caplog):
        with caplog.at_level(level=logging.INFO, logger=CRED_PROVIDER_LOGGER_NAME):
            caplog.clear()
            result = get_credentials_and_project_id()
        mock_auth_default.assert_called_once_with(scopes=None)
        assert result == ("CREDENTIALS", "PROJECT_ID")
        assert (
            "Getting connection using `google.auth.default()` since no explicit credentials are provided."
        ) in caplog.messages

    @mock.patch("google.auth.default")
    def test_get_credentials_and_project_id_with_default_auth_and_delegate(self, mock_auth_default):
        mock_credentials = mock.MagicMock()
        mock_auth_default.return_value = (mock_credentials, self.test_project_id)

        result = get_credentials_and_project_id(delegate_to="USER")
        mock_auth_default.assert_called_once_with(scopes=None)
        mock_credentials.with_subject.assert_called_once_with("USER")
        assert (mock_credentials.with_subject.return_value, self.test_project_id) == result

    @pytest.mark.parametrize("scopes", [["scope1"], ["scope1", "scope2"]])
    @mock.patch("google.auth.default")
    def test_get_credentials_and_project_id_with_default_auth_and_scopes(self, mock_auth_default, scopes):
        mock_credentials = mock.MagicMock()
        mock_auth_default.return_value = (mock_credentials, self.test_project_id)

        result = get_credentials_and_project_id(scopes=scopes)
        mock_auth_default.assert_called_once_with(scopes=scopes)
        assert mock_auth_default.return_value == result

    @mock.patch(
        "airflow.providers.google.cloud.utils.credentials_provider.impersonated_credentials.Credentials"
    )
    @mock.patch("google.auth.default")
    def test_get_credentials_and_project_id_with_default_auth_and_target_principal(
        self, mock_auth_default, mock_impersonated_credentials
    ):
        mock_credentials = mock.MagicMock()
        mock_auth_default.return_value = (mock_credentials, self.test_project_id)

        result = get_credentials_and_project_id(
            target_principal=ACCOUNT_3_ANOTHER_PROJECT,
        )
        mock_auth_default.assert_called_once_with(scopes=None)
        mock_impersonated_credentials.assert_called_once_with(
            source_credentials=mock_credentials,
            target_principal=ACCOUNT_3_ANOTHER_PROJECT,
            delegates=None,
            target_scopes=None,
        )
        assert (mock_impersonated_credentials.return_value, ANOTHER_PROJECT_ID) == result

    @mock.patch(
        "airflow.providers.google.cloud.utils.credentials_provider.impersonated_credentials.Credentials"
    )
    @mock.patch("google.auth.default")
    def test_get_credentials_and_project_id_with_default_auth_and_scopes_and_target_principal(
        self, mock_auth_default, mock_impersonated_credentials
    ):
        mock_credentials = mock.MagicMock()
        mock_auth_default.return_value = (mock_credentials, self.test_project_id)

        result = get_credentials_and_project_id(
            scopes=["scope1", "scope2"],
            target_principal=ACCOUNT_1_SAME_PROJECT,
        )
        mock_auth_default.assert_called_once_with(scopes=["scope1", "scope2"])
        mock_impersonated_credentials.assert_called_once_with(
            source_credentials=mock_credentials,
            target_principal=ACCOUNT_1_SAME_PROJECT,
            delegates=None,
            target_scopes=["scope1", "scope2"],
        )
        assert (mock_impersonated_credentials.return_value, self.test_project_id) == result

    @mock.patch(
        "airflow.providers.google.cloud.utils.credentials_provider.impersonated_credentials.Credentials"
    )
    @mock.patch("google.auth.default")
    def test_get_credentials_and_project_id_with_default_auth_and_target_principal_and_delegates(
        self, mock_auth_default, mock_impersonated_credentials
    ):
        mock_credentials = mock.MagicMock()
        mock_auth_default.return_value = (mock_credentials, self.test_project_id)

        result = get_credentials_and_project_id(
            target_principal=ACCOUNT_3_ANOTHER_PROJECT,
            delegates=[ACCOUNT_1_SAME_PROJECT, ACCOUNT_2_SAME_PROJECT],
        )
        mock_auth_default.assert_called_once_with(scopes=None)
        mock_impersonated_credentials.assert_called_once_with(
            source_credentials=mock_credentials,
            target_principal=ACCOUNT_3_ANOTHER_PROJECT,
            delegates=[ACCOUNT_1_SAME_PROJECT, ACCOUNT_2_SAME_PROJECT],
            target_scopes=None,
        )
        assert (mock_impersonated_credentials.return_value, ANOTHER_PROJECT_ID) == result

    @mock.patch(
        "google.oauth2.service_account.Credentials.from_service_account_file",
    )
    def test_get_credentials_and_project_id_with_service_account_file(
        self, mock_from_service_account_file, caplog
    ):
        mock_from_service_account_file.return_value.project_id = self.test_project_id
        with caplog.at_level(level=logging.DEBUG, logger=CRED_PROVIDER_LOGGER_NAME):
            caplog.clear()
            result = get_credentials_and_project_id(key_path=self.test_key_file)
        mock_from_service_account_file.assert_called_once_with(self.test_key_file, scopes=None)
        assert (mock_from_service_account_file.return_value, self.test_project_id) == result
        assert "Getting connection using JSON key file KEY_PATH.json" in caplog.messages

    @pytest.mark.parametrize(
        "file", [pytest.param("path/to/file.p12", id="p12"), pytest.param("incorrect_file.ext", id="unknown")]
    )
    def test_get_credentials_and_project_id_with_service_account_file_and_non_valid_key(self, file):
        with pytest.raises(AirflowException):
            get_credentials_and_project_id(key_path=file)

    @mock.patch(
        "google.oauth2.service_account.Credentials.from_service_account_info",
    )
    def test_get_credentials_and_project_id_with_service_account_info(
        self, mock_from_service_account_info, caplog
    ):
        mock_from_service_account_info.return_value.project_id = self.test_project_id
        service_account = {"private_key": "PRIVATE_KEY"}
        with caplog.at_level(level=logging.DEBUG, logger=CRED_PROVIDER_LOGGER_NAME):
            caplog.clear()
            result = get_credentials_and_project_id(keyfile_dict=service_account)
        mock_from_service_account_info.assert_called_once_with(service_account, scopes=None)
        assert (mock_from_service_account_info.return_value, self.test_project_id) == result
        assert "Getting connection using JSON Dict" in caplog.messages

    @mock.patch("google.auth.load_credentials_from_file", return_value=("CREDENTIALS", "PROJECT_ID"))
    def test_get_credentials_using_credential_config_file(self, mock_load_credentials_from_file, caplog):
        with (
            caplog.at_level(level=logging.DEBUG, logger=CRED_PROVIDER_LOGGER_NAME),
            NamedTemporaryFile() as temp_file,
        ):
            caplog.clear()
            result = get_credentials_and_project_id(credential_config_file=temp_file.name)
        mock_load_credentials_from_file.assert_called_once_with(temp_file.name, scopes=None)
        assert mock_load_credentials_from_file.return_value == result
        assert (
            f"Getting connection using credential configuration file: `{temp_file.name}`" in caplog.messages
        )

    @mock.patch("google.auth.load_credentials_from_file", return_value=("CREDENTIALS", "PROJECT_ID"))
    def test_get_credentials_using_credential_config_dict(self, mock_load_credentials_from_file, caplog):
        with caplog.at_level(level=logging.DEBUG, logger=CRED_PROVIDER_LOGGER_NAME):
            caplog.clear()
            result = get_credentials_and_project_id(credential_config_file={"type": "external_account"})
        mock_load_credentials_from_file.assert_called_once()
        assert mock_load_credentials_from_file.return_value == result
        assert "Getting connection using credential configuration dict." in caplog.messages

    @mock.patch("google.auth.load_credentials_from_file", return_value=("CREDENTIALS", "PROJECT_ID"))
    def test_get_credentials_using_credential_config_string(self, mock_load_credentials_from_file, caplog):
        with caplog.at_level(level=logging.DEBUG, logger=CRED_PROVIDER_LOGGER_NAME):
            caplog.clear()
            result = get_credentials_and_project_id(credential_config_file='{"type": "external_account"}')
        mock_load_credentials_from_file.assert_called_once()
        assert mock_load_credentials_from_file.return_value == result
        assert "Getting connection using credential configuration string." in caplog.messages

    def test_get_credentials_using_credential_config_invalid_string(self, caplog):
        caplog.clear()
        with (
            pytest.raises(DefaultCredentialsError),
            caplog.at_level(level=logging.DEBUG, logger=CRED_PROVIDER_LOGGER_NAME),
        ):
            get_credentials_and_project_id(credential_config_file="invalid json}}}}")
        assert "Getting connection using credential configuration string." in caplog.messages

    @mock.patch("google.auth.default", return_value=("CREDENTIALS", "PROJECT_ID"))
    @mock.patch("google.oauth2.service_account.Credentials.from_service_account_info")
    @mock.patch("airflow.providers.google.cloud.utils.credentials_provider._SecretManagerClient")
    def test_get_credentials_and_project_id_with_key_secret_name(
        self, mock_secret_manager_client, mock_from_service_account_info, mock_default
    ):
        mock_secret_manager_client.return_value.is_valid_secret_name.return_value = True
        mock_secret_manager_client.return_value.get_secret.return_value = (
            '{"type":"service_account","project_id":"pid","private_key_id":"pkid","private_key":"payload"}'
        )

        get_credentials_and_project_id(key_secret_name="secret name")
        mock_from_service_account_info.assert_called_once_with(
            {
                "type": "service_account",
                "project_id": "pid",
                "private_key_id": "pkid",
                "private_key": "payload",
            },
            scopes=ANY,
        )

    @mock.patch("google.auth.default", return_value=("CREDENTIALS", "PROJECT_ID"))
    @mock.patch("airflow.providers.google.cloud.utils.credentials_provider._SecretManagerClient")
    def test_get_credentials_and_project_id_with_key_secret_name_when_key_is_invalid(
        self, mock_secret_manager_client, mock_default
    ):
        mock_secret_manager_client.return_value.is_valid_secret_name.return_value = True
        mock_secret_manager_client.return_value.get_secret.return_value = ""

        with pytest.raises(
            AirflowException,
            match=re.escape("Key data read from GCP Secret Manager is not valid JSON."),
        ):
            get_credentials_and_project_id(key_secret_name="secret name")

    def test_get_credentials_and_project_id_with_mutually_exclusive_configuration(self):
        with pytest.raises(AirflowException, match="mutually exclusive."):
            get_credentials_and_project_id(key_path="KEY.json", keyfile_dict={"private_key": "PRIVATE_KEY"})

    @mock.patch("airflow.providers.google.cloud.utils.credentials_provider.AnonymousCredentials")
    def test_get_credentials_using_anonymous_credentials(self, mock_anonymous_credentials):
        result = get_credentials_and_project_id(is_anonymous=True)
        assert result == (mock_anonymous_credentials.return_value, "")

    @mock.patch("google.auth.default", return_value=("CREDENTIALS", "PROJECT_ID"))
    @mock.patch(
        "google.oauth2.service_account.Credentials.from_service_account_info",
    )
    @mock.patch(
        "google.oauth2.service_account.Credentials.from_service_account_file",
    )
    def test_disable_logging(self, mock_default, mock_info, mock_file, assert_no_logs):
        """Test disable logging in ``get_credentials_and_project_id``"""

        # assert no logs
        with assert_no_logs(level="DEBUG", logger=CRED_PROVIDER_LOGGER_NAME):
            get_credentials_and_project_id(disable_logging=True)

        # assert no debug logs emitted from get_credentials_and_project_id
        with assert_no_logs(level="DEBUG", logger=CRED_PROVIDER_LOGGER_NAME):
            get_credentials_and_project_id(
                keyfile_dict={"private_key": "PRIVATE_KEY"},
                disable_logging=True,
            )

        # assert no debug logs emitted from get_credentials_and_project_id
        with assert_no_logs(level="DEBUG", logger=CRED_PROVIDER_LOGGER_NAME):
            get_credentials_and_project_id(
                key_path="KEY.json",
                disable_logging=True,
            )

    @mock.patch("google.auth.load_credentials_from_dict", return_value=("CREDENTIALS", "PROJECT_ID"))
    def test_get_credentials_using_identity_provider(self, mock_load_credentials_from_file, caplog):
        with caplog.at_level(level=logging.DEBUG, logger=CRED_PROVIDER_LOGGER_NAME):
            caplog.clear()
            result = get_credentials_and_project_id(
                credential_config_file=CREDENTIAL_CONFIG_STRING_FILE,
                idp_issuer_url=IDP_LINK,
                client_id=CLIENT_ID,
                client_secret=CLIENT_SECRET,
            )
            mock_load_credentials_from_file.assert_called_once_with(
                info={
                    "audience": TEST_AUDIENCE,
                    "subject_token_type": TOKEN_TYPE,
                    "service_account_impersonation_url": ACCOUNT_IMPERSONATION,
                    "subject_token_supplier": ANY,
                },
                scopes=ANY,
            )
            assert result == ("CREDENTIALS", "PROJECT_ID")
            assert (
                "Getting connection using credential configuration file and external Identity Provider."
                in caplog.messages
            )

    @mock.patch("google.auth.load_credentials_from_dict", return_value=("CREDENTIALS", "PROJECT_ID"))
    def test_get_credentials_using_idp_dict_config_file(self, mock_load_credentials_from_file, caplog):
        with caplog.at_level(level=logging.DEBUG, logger=CRED_PROVIDER_LOGGER_NAME):
            caplog.clear()
            result = get_credentials_and_project_id(
                credential_config_file=CREDENTIAL_CONFIG_DICT_FILE,
                idp_issuer_url=IDP_LINK,
                client_id=CLIENT_ID,
                client_secret=CLIENT_SECRET,
            )
            mock_load_credentials_from_file.assert_called_once_with(
                info={
                    "audience": TEST_AUDIENCE,
                    "subject_token_type": TOKEN_TYPE,
                    "service_account_impersonation_url": ACCOUNT_IMPERSONATION,
                    "subject_token_supplier": ANY,
                },
                scopes=ANY,
            )
            assert result == ("CREDENTIALS", "PROJECT_ID")
            assert (
                "Getting connection using credential configuration file and external Identity Provider."
                in caplog.messages
            )

    def test_get_credentials_using_idp_no_credential_config(self):
        with pytest.raises(
            AirflowException,
            match=re.escape(
                "Credential Configuration File is needed to use authentication by External Identity Provider."
            ),
        ):
            get_credentials_and_project_id(
                idp_issuer_url=IDP_LINK,
                client_id=CLIENT_ID,
                client_secret=CLIENT_SECRET,
            )

    def test_get_credentials_using_idp_invalid_json_credential_config(self):
        with pytest.raises(
            AirflowException,
            match=re.escape("Credential Configuration File is not a valid json string."),
        ):
            get_credentials_and_project_id(
                credential_config_file="invalid json}}}}",
                idp_issuer_url=IDP_LINK,
                client_id=CLIENT_ID,
                client_secret=CLIENT_SECRET,
            )

    def test_get_credentials_using_idp_invalid_format_credential_config(self):
        with pytest.raises(
            AirflowException,
            match=re.escape(f"Invalid argument type, expected str or dict, got {bool}."),
        ):
            get_credentials_and_project_id(
                credential_config_file=True,
                idp_issuer_url=IDP_LINK,
                client_id=CLIENT_ID,
                client_secret=CLIENT_SECRET,
            )


class TestGetScopes:
    def test_get_scopes_with_default(self):
        assert _get_scopes() == _DEFAULT_SCOPES

    @pytest.mark.parametrize(
        ("scopes_str", "scopes"),
        [
            pytest.param("scope1", ["scope1"], id="single-scope"),
            pytest.param("scope1,scope2", ["scope1", "scope2"], id="multiple-scopes"),
        ],
    )
    def test_get_scopes_with_input(self, scopes_str, scopes):
        assert _get_scopes(scopes_str) == scopes


class TestGetTargetPrincipalAndDelegates:
    def test_get_target_principal_and_delegates_no_argument(self):
        assert _get_target_principal_and_delegates() == (None, None)

    @pytest.mark.parametrize(
        ("impersonation_chain", "target_principal_and_delegates"),
        [
            pytest.param(ACCOUNT_1_SAME_PROJECT, (ACCOUNT_1_SAME_PROJECT, None), id="string"),
            pytest.param([], (None, None), id="empty-list"),
            pytest.param([ACCOUNT_1_SAME_PROJECT], (ACCOUNT_1_SAME_PROJECT, []), id="single-element-list"),
            pytest.param(
                [ACCOUNT_1_SAME_PROJECT, ACCOUNT_2_SAME_PROJECT, ACCOUNT_3_ANOTHER_PROJECT],
                (ACCOUNT_3_ANOTHER_PROJECT, [ACCOUNT_1_SAME_PROJECT, ACCOUNT_2_SAME_PROJECT]),
                id="multiple-elements-list",
            ),
        ],
    )
    def test_get_target_principal_and_delegates_with_input(
        self, impersonation_chain, target_principal_and_delegates
    ):
        assert _get_target_principal_and_delegates(impersonation_chain) == target_principal_and_delegates


class TestGetProjectIdFromServiceAccountEmail:
    def test_get_project_id_from_service_account_email(self):
        assert _get_project_id_from_service_account_email(ACCOUNT_3_ANOTHER_PROJECT) == ANOTHER_PROJECT_ID

    def test_get_project_id_from_service_account_email_wrong_input(self):
        with pytest.raises(AirflowException):
            _get_project_id_from_service_account_email("ACCOUNT_1")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import MagicMock

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.transfers.gcs_to_local import MAX_XCOM_SIZE, GCSToLocalFilesystemOperator

TASK_ID = "test-gcs-operator"
TEST_BUCKET = "test-bucket"
TEST_PROJECT = "test-project"
DELIMITER = ".csv"
PREFIX = "TEST"
MOCK_FILES = ["TEST1.csv", "TEST2.csv", "TEST3.csv"]
TEST_OBJECT = "dir1/test-object"
LOCAL_FILE_PATH = "/home/airflow/gcp/test-object"
XCOM_KEY = "some_xkom_key"
FILE_CONTENT_STR = "some file content"
FILE_CONTENT_BYTES_UTF8 = b"some file content"
FILE_CONTENT_BYTES_UTF16 = (
    b"\xff\xfes\x00o\x00m\x00e\x00 \x00f\x00i\x00l\x00e\x00 \x00c\x00o\x00n\x00t\x00e\x00n\x00t\x00"
)


class TestGoogleCloudStorageDownloadOperator:
    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_local.GCSHook")
    def test_execute(self, mock_hook):
        operator = GCSToLocalFilesystemOperator(
            task_id=TASK_ID,
            bucket=TEST_BUCKET,
            object_name=TEST_OBJECT,
            filename=LOCAL_FILE_PATH,
        )

        operator.execute(None)
        mock_hook.return_value.download.assert_called_once_with(
            bucket_name=TEST_BUCKET, object_name=TEST_OBJECT, filename=LOCAL_FILE_PATH
        )

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_local.GCSHook")
    def test_size_lt_max_xcom_size(self, mock_hook):
        operator = GCSToLocalFilesystemOperator(
            task_id=TASK_ID,
            bucket=TEST_BUCKET,
            object_name=TEST_OBJECT,
            store_to_xcom_key=XCOM_KEY,
        )
        context = {"ti": MagicMock()}
        mock_hook.return_value.download.return_value = FILE_CONTENT_BYTES_UTF8
        mock_hook.return_value.get_size.return_value = MAX_XCOM_SIZE - 1

        operator.execute(context=context)
        mock_hook.return_value.get_size.assert_called_once_with(
            bucket_name=TEST_BUCKET, object_name=TEST_OBJECT
        )
        mock_hook.return_value.download.assert_called_once_with(
            bucket_name=TEST_BUCKET, object_name=TEST_OBJECT
        )
        context["ti"].xcom_push.assert_called_once_with(key=XCOM_KEY, value=FILE_CONTENT_STR)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_local.GCSHook")
    def test_size_gt_max_xcom_size(self, mock_hook):
        operator = GCSToLocalFilesystemOperator(
            task_id=TASK_ID,
            bucket=TEST_BUCKET,
            object_name=TEST_OBJECT,
            store_to_xcom_key=XCOM_KEY,
        )
        context = {"ti": MagicMock()}
        mock_hook.return_value.download.return_value = FILE_CONTENT_BYTES_UTF8
        mock_hook.return_value.get_size.return_value = MAX_XCOM_SIZE + 1

        with pytest.raises(AirflowException, match="file is too large"):
            operator.execute(context=context)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_local.GCSHook")
    def test_xcom_encoding(self, mock_hook):
        operator = GCSToLocalFilesystemOperator(
            task_id=TASK_ID,
            bucket=TEST_BUCKET,
            object_name=TEST_OBJECT,
            store_to_xcom_key=XCOM_KEY,
            file_encoding="utf-16",
        )
        context = {"ti": MagicMock()}
        mock_hook.return_value.download.return_value = FILE_CONTENT_BYTES_UTF16
        mock_hook.return_value.get_size.return_value = MAX_XCOM_SIZE - 1

        operator.execute(context=context)
        mock_hook.return_value.get_size.assert_called_once_with(
            bucket_name=TEST_BUCKET, object_name=TEST_OBJECT
        )
        mock_hook.return_value.download.assert_called_once_with(
            bucket_name=TEST_BUCKET, object_name=TEST_OBJECT
        )
        context["ti"].xcom_push.assert_called_once_with(key=XCOM_KEY, value=FILE_CONTENT_STR)

    def test_get_openlineage_facets_on_start_(self):
        operator = GCSToLocalFilesystemOperator(
            task_id=TASK_ID,
            bucket=TEST_BUCKET,
            object_name=TEST_OBJECT,
            filename=LOCAL_FILE_PATH,
        )
        result = operator.get_openlineage_facets_on_start()
        assert not result.job_facets
        assert not result.run_facets
        assert len(result.outputs) == 1
        assert len(result.inputs) == 1
        assert result.outputs[0].namespace == "file"
        assert result.outputs[0].name == LOCAL_FILE_PATH
        assert result.inputs[0].namespace == f"gs://{TEST_BUCKET}"
        assert result.inputs[0].name == TEST_OBJECT



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google DataFusion operators."""

from __future__ import annotations

import time
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from google.api_core.retry import exponential_sleep_generator
from googleapiclient.errors import HttpError

from airflow.configuration import conf
from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.datafusion import SUCCESS_STATES, DataFusionHook, PipelineStates
from airflow.providers.google.cloud.links.datafusion import (
    DataFusionInstanceLink,
    DataFusionPipelineLink,
    DataFusionPipelinesLink,
)
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator
from airflow.providers.google.cloud.triggers.datafusion import DataFusionStartPipelineTrigger
from airflow.providers.google.cloud.utils.datafusion import DataFusionPipelineType
from airflow.providers.google.cloud.utils.helpers import resource_path_to_dict
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context
    from airflow.providers.openlineage.extractors import OperatorLineage


class CloudDataFusionRestartInstanceOperator(GoogleCloudBaseOperator):
    """
    Restart a single Data Fusion instance.

    At the end of an operation instance is fully restarted.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDataFusionRestartInstanceOperator`

    :param instance_name: The name of the instance to restart.
    :param location: The Cloud Data Fusion location in which to handle the request.
    :param project_id: The ID of the Google Cloud project that the instance belongs to.
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "instance_name",
        "impersonation_chain",
    )
    operator_extra_links = (DataFusionInstanceLink(),)

    def __init__(
        self,
        *,
        instance_name: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        api_version: str = "v1beta1",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.instance_name = instance_name
        self.location = location
        self.project_id = project_id
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        hook = DataFusionHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Restarting Data Fusion instance: %s", self.instance_name)
        operation = hook.restart_instance(
            instance_name=self.instance_name,
            location=self.location,
            project_id=self.project_id,
        )
        instance = hook.wait_for_operation(operation)
        self.log.info("Instance %s restarted successfully", self.instance_name)

        project_id = resource_path_to_dict(resource_name=instance["name"])["projects"]
        DataFusionInstanceLink.persist(
            context=context,
            project_id=project_id,
            instance_name=self.instance_name,
            region=self.location,
        )


class CloudDataFusionDeleteInstanceOperator(GoogleCloudBaseOperator):
    """
    Deletes a single Date Fusion instance.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDataFusionDeleteInstanceOperator`

    :param instance_name: The name of the instance to restart.
    :param location: The Cloud Data Fusion location in which to handle the request.
    :param project_id: The ID of the Google Cloud project that the instance belongs to.
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "instance_name",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        instance_name: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        api_version: str = "v1beta1",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.instance_name = instance_name
        self.location = location
        self.project_id = project_id
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        hook = DataFusionHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Deleting Data Fusion instance: %s", self.instance_name)
        operation = hook.delete_instance(
            instance_name=self.instance_name,
            location=self.location,
            project_id=self.project_id,
        )
        hook.wait_for_operation(operation)
        self.log.info("Instance %s deleted successfully", self.instance_name)


class CloudDataFusionCreateInstanceOperator(GoogleCloudBaseOperator):
    """
    Creates a new Data Fusion instance in the specified project and location.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDataFusionCreateInstanceOperator`

    :param instance_name: The name of the instance to create.
    :param instance: An instance of Instance.
        https://cloud.google.com/data-fusion/docs/reference/rest/v1beta1/projects.locations.instances#Instance
    :param location: The Cloud Data Fusion location in which to handle the request.
    :param project_id: The ID of the Google Cloud project that the instance belongs to.
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "instance_name",
        "instance",
        "impersonation_chain",
    )
    operator_extra_links = (DataFusionInstanceLink(),)

    def __init__(
        self,
        *,
        instance_name: str,
        instance: dict[str, Any],
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        api_version: str = "v1beta1",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.instance_name = instance_name
        self.instance = instance
        self.location = location
        self.project_id = project_id
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> dict:
        hook = DataFusionHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Creating Data Fusion instance: %s", self.instance_name)
        try:
            operation = hook.create_instance(
                instance_name=self.instance_name,
                instance=self.instance,
                location=self.location,
                project_id=self.project_id,
            )
            instance = hook.wait_for_operation(operation)
            self.log.info("Instance %s created successfully", self.instance_name)
        except HttpError as err:
            if err.resp.status not in (409, "409"):
                raise
            self.log.info("Instance %s already exists", self.instance_name)
            instance = hook.get_instance(
                instance_name=self.instance_name, location=self.location, project_id=self.project_id
            )
            # Wait for instance to be ready
            for time_to_wait in exponential_sleep_generator(initial=10, maximum=120):
                if instance["state"] != "CREATING":
                    break
                time.sleep(time_to_wait)
                instance = hook.get_instance(
                    instance_name=self.instance_name, location=self.location, project_id=self.project_id
                )

        project_id = resource_path_to_dict(resource_name=instance["name"])["projects"]
        DataFusionInstanceLink.persist(
            context=context,
            project_id=project_id,
            instance_name=self.instance_name,
            region=self.location,
        )
        return instance


class CloudDataFusionUpdateInstanceOperator(GoogleCloudBaseOperator):
    """
    Updates a single Data Fusion instance.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDataFusionUpdateInstanceOperator`

    :param instance_name: The name of the instance to create.
    :param instance: An instance of Instance.
        https://cloud.google.com/data-fusion/docs/reference/rest/v1beta1/projects.locations.instances#Instance
    :param update_mask: Field mask is used to specify the fields that the update will overwrite
        in an instance resource. The fields specified in the updateMask are relative to the resource,
        not the full request. A field will be overwritten if it is in the mask. If the user does not
        provide a mask, all the supported fields (labels and options currently) will be overwritten.
        A comma-separated list of fully qualified names of fields. Example: "user.displayName,photo".
        https://developers.google.com/protocol-buffers/docs/reference/google.protobuf?_ga=2.205612571.-968688242.1573564810#google.protobuf.FieldMask
    :param location: The Cloud Data Fusion location in which to handle the request.
    :param project_id: The ID of the Google Cloud project that the instance belongs to.
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "instance_name",
        "instance",
        "impersonation_chain",
    )
    operator_extra_links = (DataFusionInstanceLink(),)

    def __init__(
        self,
        *,
        instance_name: str,
        instance: dict[str, Any],
        update_mask: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        api_version: str = "v1beta1",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.update_mask = update_mask
        self.instance_name = instance_name
        self.instance = instance
        self.location = location
        self.project_id = project_id
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        hook = DataFusionHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Updating Data Fusion instance: %s", self.instance_name)
        operation = hook.patch_instance(
            instance_name=self.instance_name,
            instance=self.instance,
            update_mask=self.update_mask,
            location=self.location,
            project_id=self.project_id,
        )
        instance = hook.wait_for_operation(operation)
        self.log.info("Instance %s updated successfully", self.instance_name)

        project_id = resource_path_to_dict(resource_name=instance["name"])["projects"]
        DataFusionInstanceLink.persist(
            context=context,
            project_id=project_id,
            instance_name=self.instance_name,
            region=self.location,
        )


class CloudDataFusionGetInstanceOperator(GoogleCloudBaseOperator):
    """
    Gets details of a single Data Fusion instance.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDataFusionGetInstanceOperator`

    :param instance_name: The name of the instance.
    :param location: The Cloud Data Fusion location in which to handle the request.
    :param project_id: The ID of the Google Cloud project that the instance belongs to.
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "instance_name",
        "impersonation_chain",
    )
    operator_extra_links = (DataFusionInstanceLink(),)

    def __init__(
        self,
        *,
        instance_name: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        api_version: str = "v1beta1",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.instance_name = instance_name
        self.location = location
        self.project_id = project_id
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> dict:
        hook = DataFusionHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Retrieving Data Fusion instance: %s", self.instance_name)
        instance = hook.get_instance(
            instance_name=self.instance_name,
            location=self.location,
            project_id=self.project_id,
        )

        project_id = resource_path_to_dict(resource_name=instance["name"])["projects"]
        DataFusionInstanceLink.persist(
            context=context,
            project_id=project_id,
            instance_name=self.instance_name,
            region=self.location,
        )
        return instance


class CloudDataFusionCreatePipelineOperator(GoogleCloudBaseOperator):
    """
    Creates a Cloud Data Fusion pipeline.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDataFusionCreatePipelineOperator`

    :param pipeline_name: Your pipeline name.
    :param pipeline: The pipeline definition. For more information check:
        https://docs.cdap.io/cdap/current/en/developer-manual/pipelines/developing-pipelines.html#pipeline-configuration-file-format
    :param instance_name: The name of the instance.
    :param location: The Cloud Data Fusion location in which to handle the request.
    :param namespace: If your pipeline belongs to a Basic edition instance, the namespace ID
        is always default. If your pipeline belongs to an Enterprise edition instance, you
        can create a namespace.
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    operator_extra_links = (DataFusionPipelineLink(),)

    template_fields: Sequence[str] = (
        "instance_name",
        "pipeline_name",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        pipeline_name: str,
        pipeline: dict[str, Any],
        instance_name: str,
        location: str,
        namespace: str = "default",
        project_id: str = PROVIDE_PROJECT_ID,
        api_version: str = "v1beta1",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.pipeline_name = pipeline_name
        self.pipeline = pipeline
        self.namespace = namespace
        self.instance_name = instance_name
        self.location = location
        self.project_id = project_id
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        hook = DataFusionHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Creating Data Fusion pipeline: %s", self.pipeline_name)
        instance = hook.get_instance(
            instance_name=self.instance_name,
            location=self.location,
            project_id=self.project_id,
        )
        api_url = instance["apiEndpoint"]
        hook.create_pipeline(
            pipeline_name=self.pipeline_name,
            pipeline=self.pipeline,
            instance_url=api_url,
            namespace=self.namespace,
        )
        DataFusionPipelineLink.persist(
            context=context,
            uri=instance["serviceEndpoint"],
            pipeline_name=self.pipeline_name,
            namespace=self.namespace,
        )
        self.log.info("Pipeline %s created", self.pipeline_name)


class CloudDataFusionDeletePipelineOperator(GoogleCloudBaseOperator):
    """
    Deletes a Cloud Data Fusion pipeline.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDataFusionDeletePipelineOperator`

    :param pipeline_name: Your pipeline name.
    :param version_id: Version of pipeline to delete
    :param instance_name: The name of the instance.
    :param location: The Cloud Data Fusion location in which to handle the request.
    :param namespace: If your pipeline belongs to a Basic edition instance, the namespace ID
        is always default. If your pipeline belongs to an Enterprise edition instance, you
        can create a namespace.
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "instance_name",
        "version_id",
        "pipeline_name",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        pipeline_name: str,
        instance_name: str,
        location: str,
        version_id: str | None = None,
        namespace: str = "default",
        project_id: str = PROVIDE_PROJECT_ID,
        api_version: str = "v1beta1",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.pipeline_name = pipeline_name
        self.version_id = version_id
        self.namespace = namespace
        self.instance_name = instance_name
        self.location = location
        self.project_id = project_id
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        hook = DataFusionHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Deleting Data Fusion pipeline: %s", self.pipeline_name)
        instance = hook.get_instance(
            instance_name=self.instance_name,
            location=self.location,
            project_id=self.project_id,
        )
        api_url = instance["apiEndpoint"]
        hook.delete_pipeline(
            pipeline_name=self.pipeline_name,
            version_id=self.version_id,
            instance_url=api_url,
            namespace=self.namespace,
        )
        self.log.info("Pipeline deleted")


class CloudDataFusionListPipelinesOperator(GoogleCloudBaseOperator):
    """
    Lists Cloud Data Fusion pipelines.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDataFusionListPipelinesOperator`


    :param instance_name: The name of the instance.
    :param location: The Cloud Data Fusion location in which to handle the request.
    :param artifact_version: Artifact version to filter instances
    :param artifact_name: Artifact name to filter instances
    :param namespace: If your pipeline belongs to a Basic edition instance, the namespace ID
        is always default. If your pipeline belongs to an Enterprise edition instance, you
        can create a namespace.
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "instance_name",
        "artifact_name",
        "artifact_version",
        "impersonation_chain",
    )
    operator_extra_links = (DataFusionPipelinesLink(),)

    def __init__(
        self,
        *,
        instance_name: str,
        location: str,
        artifact_name: str | None = None,
        artifact_version: str | None = None,
        namespace: str = "default",
        project_id: str = PROVIDE_PROJECT_ID,
        api_version: str = "v1beta1",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.artifact_version = artifact_version
        self.artifact_name = artifact_name
        self.namespace = namespace
        self.instance_name = instance_name
        self.location = location
        self.project_id = project_id
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> dict:
        hook = DataFusionHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Listing Data Fusion pipelines")
        instance = hook.get_instance(
            instance_name=self.instance_name,
            location=self.location,
            project_id=self.project_id,
        )
        api_url = instance["apiEndpoint"]
        service_endpoint = instance["serviceEndpoint"]
        pipelines = hook.list_pipelines(
            instance_url=api_url,
            namespace=self.namespace,
            artifact_version=self.artifact_version,
            artifact_name=self.artifact_name,
        )
        self.log.info("Pipelines: %s", pipelines)

        DataFusionPipelinesLink.persist(
            context=context,
            uri=service_endpoint,
            namespace=self.namespace,
        )
        return pipelines


class CloudDataFusionStartPipelineOperator(GoogleCloudBaseOperator):
    """
    Starts a Cloud Data Fusion pipeline. Works for both batch and stream pipelines.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDataFusionStartPipelineOperator`

    :param pipeline_name: Your pipeline name.
    :param pipeline_type: Optional pipeline type (BATCH by default).
    :param instance_name: The name of the instance.
    :param success_states: If provided the operator will wait for pipeline to be in one of
        the provided states.
    :param pipeline_timeout: How long (in seconds) operator should wait for the pipeline to be in one of
        ``success_states``. Works only if ``success_states`` are provided.
    :param location: The Cloud Data Fusion location in which to handle the request.
    :param runtime_args: Optional runtime args to be passed to the pipeline
    :param namespace: If your pipeline belongs to a Basic edition instance, the namespace ID
        is always default. If your pipeline belongs to an Enterprise edition instance, you
        can create a namespace.
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param asynchronous: Flag to return after submitting the pipeline ID to the Data Fusion API.
        This is useful for submitting long-running pipelines and
        waiting on them asynchronously using the CloudDataFusionPipelineStateSensor
    :param deferrable: Run operator in the deferrable mode. Is not related to asynchronous parameter. While
        asynchronous parameter gives a possibility to wait until pipeline reaches terminate state using
        sleep() method, deferrable mode checks for the state using asynchronous calls. It is not possible to
        use both asynchronous and deferrable parameters at the same time.
    :param poll_interval: Polling period in seconds to check for the status. Used only in deferrable mode.
    """

    template_fields: Sequence[str] = (
        "instance_name",
        "pipeline_name",
        "runtime_args",
        "impersonation_chain",
    )
    operator_extra_links = (DataFusionPipelineLink(),)

    def __init__(
        self,
        *,
        pipeline_name: str,
        instance_name: str,
        location: str,
        pipeline_type: DataFusionPipelineType = DataFusionPipelineType.BATCH,
        runtime_args: dict[str, Any] | None = None,
        success_states: list[str] | None = None,
        namespace: str = "default",
        pipeline_timeout: int = 5 * 60,
        project_id: str = PROVIDE_PROJECT_ID,
        api_version: str = "v1beta1",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        asynchronous=False,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poll_interval=3.0,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.pipeline_name = pipeline_name
        self.pipeline_type = pipeline_type
        self.runtime_args = runtime_args
        self.namespace = namespace
        self.instance_name = instance_name
        self.location = location
        self.project_id = project_id
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.asynchronous = asynchronous
        self.pipeline_timeout = pipeline_timeout
        self.deferrable = deferrable
        self.poll_interval = poll_interval
        self.pipeline_id: str | None = None

        if success_states:
            self.success_states = success_states
        else:
            self.success_states = [*SUCCESS_STATES, PipelineStates.RUNNING]

    def execute(self, context: Context) -> str:
        hook = DataFusionHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Starting Data Fusion pipeline: %s", self.pipeline_name)
        instance = hook.get_instance(
            instance_name=self.instance_name,
            location=self.location,
            project_id=self.project_id,
        )
        api_url = instance["apiEndpoint"]
        self.pipeline_id = hook.start_pipeline(
            pipeline_name=self.pipeline_name,
            pipeline_type=self.pipeline_type,
            instance_url=api_url,
            namespace=self.namespace,
            runtime_args=self.runtime_args,
        )
        self.log.info("Pipeline %s submitted successfully.", self.pipeline_id)

        DataFusionPipelineLink.persist(
            context=context,
            uri=instance["serviceEndpoint"],
            pipeline_name=self.pipeline_name,
            namespace=self.namespace,
        )

        if self.deferrable:
            if self.asynchronous:
                raise AirflowException(
                    "Both asynchronous and deferrable parameters were passed. Please, provide only one."
                )
            self.defer(
                trigger=DataFusionStartPipelineTrigger(
                    success_states=self.success_states,
                    instance_url=api_url,
                    namespace=self.namespace,
                    pipeline_name=self.pipeline_name,
                    pipeline_type=self.pipeline_type.value,
                    pipeline_id=self.pipeline_id,
                    poll_interval=self.poll_interval,
                    gcp_conn_id=self.gcp_conn_id,
                    impersonation_chain=self.impersonation_chain,
                ),
                method_name="execute_complete",
            )
        else:
            if not self.asynchronous:
                # when NOT using asynchronous mode it will just wait for pipeline to finish and print message
                self.log.info(
                    "Waiting when pipeline %s will be in one of the success states", self.pipeline_id
                )
                hook.wait_for_pipeline_state(
                    success_states=self.success_states,
                    pipeline_id=self.pipeline_id,
                    pipeline_name=self.pipeline_name,
                    pipeline_type=self.pipeline_type,
                    namespace=self.namespace,
                    instance_url=api_url,
                    timeout=self.pipeline_timeout,
                )
                self.log.info("Pipeline %s discovered success state.", self.pipeline_id)
            #  otherwise, return pipeline_id so that sensor can use it later to check the pipeline state
        return self.pipeline_id

    def execute_complete(self, context: Context, event: dict[str, Any]):
        """
        Act as a callback for when the trigger fires - returns immediately.

        Relies on trigger to throw an exception, otherwise it assumes execution was successful.
        """
        if event["status"] == "error":
            raise AirflowException(event["message"])
        self.log.info(
            "%s completed with response %s ",
            self.task_id,
            event["message"],
        )
        return event["pipeline_id"]

    def get_openlineage_facets_on_complete(self, task_instance) -> OperatorLineage | None:
        """Build and return OpenLineage facets and datasets for the completed pipeline start."""
        from airflow.providers.common.compat.openlineage.facet import Dataset
        from airflow.providers.google.cloud.openlineage.facets import DataFusionRunFacet
        from airflow.providers.openlineage.extractors import OperatorLineage

        pipeline_resource = f"{self.project_id}:{self.location}:{self.instance_name}:{self.pipeline_name}"

        inputs = [Dataset(namespace="datafusion", name=pipeline_resource)]

        if self.pipeline_id:
            output_name = f"{pipeline_resource}:{self.pipeline_id}"
        else:
            output_name = f"{pipeline_resource}:unknown"
        outputs = [Dataset(namespace="datafusion", name=output_name)]

        run_facets = {
            "dataFusionRun": DataFusionRunFacet(
                runId=self.pipeline_id,
                runtimeArgs=self.runtime_args,
            )
        }

        return OperatorLineage(inputs=inputs, outputs=outputs, run_facets=run_facets, job_facets={})


class CloudDataFusionStopPipelineOperator(GoogleCloudBaseOperator):
    """
    Stops a Cloud Data Fusion pipeline. Works for both batch and stream pipelines.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDataFusionStopPipelineOperator`

    :param pipeline_name: Your pipeline name.
    :param instance_name: The name of the instance.
    :param location: The Cloud Data Fusion location in which to handle the request.
    :param namespace: If your pipeline belongs to a Basic edition instance, the namespace ID
        is always default. If your pipeline belongs to an Enterprise edition instance, you
        can create a namespace.
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "instance_name",
        "pipeline_name",
        "impersonation_chain",
    )
    operator_extra_links = (DataFusionPipelineLink(),)

    def __init__(
        self,
        *,
        pipeline_name: str,
        instance_name: str,
        location: str,
        namespace: str = "default",
        project_id: str = PROVIDE_PROJECT_ID,
        api_version: str = "v1beta1",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.pipeline_name = pipeline_name
        self.namespace = namespace
        self.instance_name = instance_name
        self.location = location
        self.project_id = project_id
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        hook = DataFusionHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Data Fusion pipeline: %s is going to be stopped", self.pipeline_name)
        instance = hook.get_instance(
            instance_name=self.instance_name,
            location=self.location,
            project_id=self.project_id,
        )
        api_url = instance["apiEndpoint"]

        DataFusionPipelineLink.persist(
            context=context,
            uri=instance["serviceEndpoint"],
            pipeline_name=self.pipeline_name,
            namespace=self.namespace,
        )
        hook.stop_pipeline(
            pipeline_name=self.pipeline_name,
            instance_url=api_url,
            namespace=self.namespace,
        )
        self.log.info("Pipeline stopped")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import logging
from copy import deepcopy
from unittest import mock
from unittest.mock import PropertyMock

import httplib2
import pytest
from aiohttp import ClientResponse
from aiohttp.helpers import TimerNoop
from googleapiclient.errors import HttpError
from yarl import URL

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks import mlengine as hook
from airflow.providers.google.cloud.hooks.mlengine import MLEngineAsyncHook

from unit.google.cloud.utils.base_gcp_mock import (
    GCP_PROJECT_ID_HOOK_UNIT_TEST,
    mock_base_gcp_hook_default_project_id,
)

mlengine_hook = MLEngineAsyncHook()
PROJECT_ID = "test-project"
JOB_ID = "test-job-id"


@pytest.mark.db_test
class TestMLEngineHook:
    def setup_method(self):
        self.hook = hook.MLEngineHook()

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook._authorize")
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.build")
    def test_mle_engine_client_creation(self, mock_build, mock_authorize):
        result = self.hook.get_conn()

        assert mock_build.return_value == result
        mock_build.assert_called_with("ml", "v1", http=mock_authorize.return_value, cache_discovery=False)

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_create_version(self, mock_get_conn):
        project_id = "test-project"
        model_name = "test-model"
        version_name = "test-version"
        version = {"name": version_name, "labels": {"other-label": "test-value"}}
        version_with_airflow_version = {
            "name": "test-version",
            "labels": {"other-label": "test-value", "airflow-version": hook._AIRFLOW_VERSION},
        }
        operation_path = f"projects/{project_id}/operations/test-operation"
        model_path = f"projects/{project_id}/models/{model_name}"
        operation_done = {"name": operation_path, "done": True}
        (
            mock_get_conn.return_value.projects.return_value.models.return_value.versions.return_value.create.return_value.execute.return_value
        ) = version
        (
            mock_get_conn.return_value.projects.return_value.operations.return_value.get.return_value.execute.return_value
        ) = {"name": operation_path, "done": True}
        create_version_response = self.hook.create_version(
            project_id=project_id, model_name=model_name, version_spec=deepcopy(version)
        )

        assert create_version_response == operation_done

        mock_get_conn.assert_has_calls(
            [
                mock.call()
                .projects()
                .models()
                .versions()
                .create(body=version_with_airflow_version, parent=model_path),
                mock.call().projects().models().versions().create().execute(num_retries=5),
                mock.call().projects().operations().get(name=version_name),
            ],
            any_order=True,
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_create_version_with_labels(self, mock_get_conn):
        project_id = "test-project"
        model_name = "test-model"
        version_name = "test-version"
        version = {"name": version_name}
        version_with_airflow_version = {
            "name": "test-version",
            "labels": {"airflow-version": hook._AIRFLOW_VERSION},
        }
        operation_path = f"projects/{project_id}/operations/test-operation"
        model_path = f"projects/{project_id}/models/{model_name}"
        operation_done = {"name": operation_path, "done": True}

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.versions.return_value.create.return_value.execute.return_value
        ) = version
        (
            mock_get_conn.return_value.projects.return_value.operations.return_value.get.return_value.execute.return_value
        ) = {"name": operation_path, "done": True}

        create_version_response = self.hook.create_version(
            project_id=project_id, model_name=model_name, version_spec=deepcopy(version)
        )

        assert create_version_response == operation_done

        mock_get_conn.assert_has_calls(
            [
                mock.call()
                .projects()
                .models()
                .versions()
                .create(body=version_with_airflow_version, parent=model_path),
                mock.call().projects().models().versions().create().execute(num_retries=5),
                mock.call().projects().operations().get(name=version_name),
            ],
            any_order=True,
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_set_default_version(self, mock_get_conn):
        project_id = "test-project"
        model_name = "test-model"
        version_name = "test-version"
        operation_path = f"projects/{project_id}/operations/test-operation"
        version_path = f"projects/{project_id}/models/{model_name}/versions/{version_name}"
        operation_done = {"name": operation_path, "done": True}

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.versions.return_value.setDefault.return_value.execute.return_value
        ) = operation_done

        set_default_version_response = self.hook.set_default_version(
            project_id=project_id, model_name=model_name, version_name=version_name
        )

        assert set_default_version_response == operation_done

        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().versions().setDefault(body={}, name=version_path),
                mock.call().projects().models().versions().setDefault().execute(num_retries=5),
            ],
            any_order=True,
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.time.sleep")
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_list_versions(self, mock_get_conn, mock_sleep):
        project_id = "test-project"
        model_name = "test-model"
        model_path = f"projects/{project_id}/models/{model_name}"
        version_names = [f"ver_{ix}" for ix in range(3)]
        response_bodies = [
            {"nextPageToken": f"TOKEN-{ix}", "versions": [ver]} for ix, ver in enumerate(version_names)
        ]
        response_bodies[-1].pop("nextPageToken")

        pages_requests = [mock.Mock(**{"execute.return_value": body}) for body in response_bodies]
        versions_mock = mock.Mock(
            **{"list.return_value": pages_requests[0], "list_next.side_effect": pages_requests[1:] + [None]}
        )

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.versions.return_value
        ) = versions_mock

        list_versions_response = self.hook.list_versions(project_id=project_id, model_name=model_name)

        assert list_versions_response == version_names
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().versions().list(pageSize=100, parent=model_path),
                mock.call().projects().models().versions().list().execute(num_retries=5),
            ]
            + [
                mock.call()
                .projects()
                .models()
                .versions()
                .list_next(previous_request=pages_requests[i], previous_response=response_bodies[i])
                for i in range(3)
            ],
            any_order=True,
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_delete_version(self, mock_get_conn):
        project_id = "test-project"
        model_name = "test-model"
        version_name = "test-version"
        operation_path = f"projects/{project_id}/operations/test-operation"
        version_path = f"projects/{project_id}/models/{model_name}/versions/{version_name}"
        version = {"name": operation_path}
        operation_not_done = {"name": operation_path, "done": False}
        operation_done = {"name": operation_path, "done": True}

        (
            mock_get_conn.return_value.projects.return_value.operations.return_value.get.return_value.execute.side_effect
        ) = [operation_not_done, operation_done]

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.versions.return_value.delete.return_value.execute.return_value
        ) = version

        delete_version_response = self.hook.delete_version(
            project_id=project_id, model_name=model_name, version_name=version_name
        )

        assert delete_version_response == operation_done
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().versions().delete(name=version_path),
                mock.call().projects().models().versions().delete().execute(num_retries=5),
                mock.call().projects().operations().get(name=operation_path),
                mock.call().projects().operations().get().execute(num_retries=5),
            ],
            any_order=True,
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_create_model(self, mock_get_conn):
        project_id = "test-project"
        model_name = "test-model"
        model = {
            "name": model_name,
        }
        model_with_airflow_version = {
            "name": model_name,
            "labels": {"airflow-version": hook._AIRFLOW_VERSION},
        }
        project_path = f"projects/{project_id}"

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.create.return_value.execute.return_value
        ) = model

        create_model_response = self.hook.create_model(project_id=project_id, model=deepcopy(model))

        assert create_model_response == model
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().create(body=model_with_airflow_version, parent=project_path),
                mock.call().projects().models().create().execute(num_retries=5),
            ]
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_create_model_idempotency(self, mock_get_conn):
        project_id = "test-project"
        model_name = "test-model"
        model = {
            "name": model_name,
        }
        model_with_airflow_version = {
            "name": model_name,
            "labels": {"airflow-version": hook._AIRFLOW_VERSION},
        }
        project_path = f"projects/{project_id}"

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.create.return_value.execute.side_effect
        ) = [
            HttpError(
                resp=httplib2.Response({"status": 409}),
                content=json.dumps(
                    {
                        "error": {
                            "code": 409,
                            "message": "Field: model.name Error: A model with the same name already exists.",
                            "status": "ALREADY_EXISTS",
                            "details": [
                                {
                                    "@type": "type.googleapis.com/google.rpc.BadRequest",
                                    "fieldViolations": [
                                        {
                                            "field": "model.name",
                                            "description": "A model with the same name already exists.",
                                        }
                                    ],
                                }
                            ],
                        }
                    }
                ).encode(),
            )
        ]

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.get.return_value.execute.return_value
        ) = deepcopy(model)

        create_model_response = self.hook.create_model(project_id=project_id, model=deepcopy(model))

        assert create_model_response == model
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().create(body=model_with_airflow_version, parent=project_path),
                mock.call().projects().models().create().execute(num_retries=5),
            ]
        )
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().get(name="projects/test-project/models/test-model"),
                mock.call().projects().models().get().execute(num_retries=5),
            ]
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_create_model_with_labels(self, mock_get_conn):
        project_id = "test-project"
        model_name = "test-model"
        model = {"name": model_name, "labels": {"other-label": "test-value"}}
        model_with_airflow_version = {
            "name": model_name,
            "labels": {"other-label": "test-value", "airflow-version": hook._AIRFLOW_VERSION},
        }
        project_path = f"projects/{project_id}"

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.create.return_value.execute.return_value
        ) = model

        create_model_response = self.hook.create_model(project_id=project_id, model=deepcopy(model))

        assert create_model_response == model
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().create(body=model_with_airflow_version, parent=project_path),
                mock.call().projects().models().create().execute(num_retries=5),
            ]
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_get_model(self, mock_get_conn):
        project_id = "test-project"
        model_name = "test-model"
        model = {"model": model_name}
        model_path = f"projects/{project_id}/models/{model_name}"

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.get.return_value.execute.return_value
        ) = model

        get_model_response = self.hook.get_model(project_id=project_id, model_name=model_name)

        assert get_model_response == model
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().get(name=model_path),
                mock.call().projects().models().get().execute(num_retries=5),
            ]
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_delete_model(self, mock_get_conn):
        project_id = "test-project"
        model_name = "test-model"
        model = {"model": model_name}
        model_path = f"projects/{project_id}/models/{model_name}"

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.delete.return_value.execute.return_value
        ) = model

        self.hook.delete_model(project_id=project_id, model_name=model_name)

        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().delete(name=model_path),
                mock.call().projects().models().delete().execute(num_retries=5),
            ]
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.log")
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_delete_model_when_not_exists(self, mock_get_conn, mock_log):
        project_id = "test-project"
        model_name = "test-model"
        model_path = f"projects/{project_id}/models/{model_name}"
        http_error = HttpError(
            resp=mock.MagicMock(status=404, reason="Model not found."), content=b"Model not found."
        )

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.delete.return_value.execute.side_effect
        ) = [http_error]

        self.hook.delete_model(project_id=project_id, model_name=model_name)

        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().delete(name=model_path),
                mock.call().projects().models().delete().execute(num_retries=5),
            ]
        )
        mock_log.error.assert_called_once_with("Model was not found: %s", http_error)

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.time.sleep")
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_delete_model_with_contents(self, mock_get_conn, mock_sleep):
        project_id = "test-project"
        model_name = "test-model"
        model_path = f"projects/{project_id}/models/{model_name}"
        operation_path = f"projects/{project_id}/operations/test-operation"
        operation_done = {"name": operation_path, "done": True}
        version_names = ["AAA", "BBB", "CCC"]
        versions = [
            {
                "name": f"projects/{project_id}/models/{model_name}/versions/{version_name}",
                "isDefault": i == 0,
            }
            for i, version_name in enumerate(version_names)
        ]

        (
            mock_get_conn.return_value.projects.return_value.operations.return_value.get.return_value.execute.return_value
        ) = operation_done
        (
            mock_get_conn.return_value.projects.return_value.models.return_value.versions.return_value.list.return_value.execute.return_value
        ) = {"versions": versions}
        (
            mock_get_conn.return_value.projects.return_value.models.return_value.versions.return_value.list_next.return_value
        ) = None

        self.hook.delete_model(project_id=project_id, model_name=model_name, delete_contents=True)

        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().delete(name=model_path),
                mock.call().projects().models().delete().execute(num_retries=5),
            ]
            + [
                mock.call()
                .projects()
                .models()
                .versions()
                .delete(
                    name=f"projects/{project_id}/models/{model_name}/versions/{version_name}",
                )
                for version_name in version_names
            ],
            any_order=True,
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.time.sleep")
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_create_mlengine_job(self, mock_get_conn, mock_sleep):
        project_id = "test-project"
        job_id = "test-job-id"
        project_path = f"projects/{project_id}"
        job_path = f"projects/{project_id}/jobs/{job_id}"
        new_job = {
            "jobId": job_id,
            "foo": 4815162342,
        }
        new_job_with_airflow_version = {
            "jobId": job_id,
            "foo": 4815162342,
            "labels": {"airflow-version": hook._AIRFLOW_VERSION},
        }

        job_succeeded = {
            "jobId": job_id,
            "state": "SUCCEEDED",
        }
        job_queued = {
            "jobId": job_id,
            "state": "QUEUED",
        }

        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.create.return_value.execute.return_value
        ) = job_queued
        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.get.return_value.execute.side_effect
        ) = [job_queued, job_succeeded]

        create_job_response = self.hook.create_job(project_id=project_id, job=deepcopy(new_job))

        assert create_job_response == job_succeeded
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().jobs().create(body=new_job_with_airflow_version, parent=project_path),
                mock.call().projects().jobs().get(name=job_path),
                mock.call().projects().jobs().get().execute(num_retries=5),
            ],
            any_order=True,
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.time.sleep")
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_create_mlengine_job_with_labels(self, mock_get_conn, mock_sleep):
        project_id = "test-project"
        job_id = "test-job-id"
        project_path = f"projects/{project_id}"
        job_path = f"projects/{project_id}/jobs/{job_id}"
        new_job = {"jobId": job_id, "foo": 4815162342, "labels": {"other-label": "test-value"}}
        new_job_with_airflow_version = {
            "jobId": job_id,
            "foo": 4815162342,
            "labels": {"other-label": "test-value", "airflow-version": hook._AIRFLOW_VERSION},
        }

        job_succeeded = {
            "jobId": job_id,
            "state": "SUCCEEDED",
        }
        job_queued = {
            "jobId": job_id,
            "state": "QUEUED",
        }

        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.create.return_value.execute.return_value
        ) = job_queued
        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.get.return_value.execute.side_effect
        ) = [job_queued, job_succeeded]

        create_job_response = self.hook.create_job(project_id=project_id, job=deepcopy(new_job))

        assert create_job_response == job_succeeded
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().jobs().create(body=new_job_with_airflow_version, parent=project_path),
                mock.call().projects().jobs().get(name=job_path),
                mock.call().projects().jobs().get().execute(num_retries=5),
            ],
            any_order=True,
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_create_mlengine_job_reuse_existing_job_by_default(self, mock_get_conn):
        project_id = "test-project"
        job_id = "test-job-id"
        project_path = f"projects/{project_id}"
        job_path = f"projects/{project_id}/jobs/{job_id}"
        job_succeeded = {
            "jobId": job_id,
            "foo": 4815162342,
            "state": "SUCCEEDED",
        }
        error_job_exists = HttpError(resp=mock.MagicMock(status=409), content=b"Job already exists")

        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.create.return_value.execute.side_effect
        ) = error_job_exists
        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.get.return_value.execute.return_value
        ) = job_succeeded

        create_job_response = self.hook.create_job(project_id=project_id, job=job_succeeded)

        assert create_job_response == job_succeeded
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().jobs().create(body=job_succeeded, parent=project_path),
                mock.call().projects().jobs().create().execute(num_retries=5),
                mock.call().projects().jobs().get(name=job_path),
                mock.call().projects().jobs().get().execute(num_retries=5),
            ],
            any_order=True,
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_create_mlengine_job_check_existing_job_failed(self, mock_get_conn):
        project_id = "test-project"
        job_id = "test-job-id"
        my_job = {
            "jobId": job_id,
            "foo": 4815162342,
            "state": "SUCCEEDED",
            "someInput": {"input": "someInput"},
        }
        different_job = {
            "jobId": job_id,
            "foo": 4815162342,
            "state": "SUCCEEDED",
            "someInput": {"input": "someDifferentInput"},
        }
        error_job_exists = HttpError(resp=mock.MagicMock(status=409), content=b"Job already exists")

        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.create.return_value.execute.side_effect
        ) = error_job_exists
        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.get.return_value.execute.return_value
        ) = different_job

        def check_input(existing_job):
            return existing_job.get("someInput") == my_job["someInput"]

        with pytest.raises(HttpError):
            self.hook.create_job(project_id=project_id, job=my_job, use_existing_job_fn=check_input)

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_create_mlengine_job_check_existing_job_success(self, mock_get_conn):
        project_id = "test-project"
        job_id = "test-job-id"
        my_job = {
            "jobId": job_id,
            "foo": 4815162342,
            "state": "SUCCEEDED",
            "someInput": {"input": "someInput"},
        }
        error_job_exists = HttpError(resp=mock.MagicMock(status=409), content=b"Job already exists")

        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.create.return_value.execute.side_effect
        ) = error_job_exists
        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.get.return_value.execute.return_value
        ) = my_job

        def check_input(existing_job):
            return existing_job.get("someInput") == my_job["someInput"]

        create_job_response = self.hook.create_job(
            project_id=project_id, job=my_job, use_existing_job_fn=check_input
        )

        assert create_job_response == my_job

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_cancel_mlengine_job(self, mock_get_conn):
        project_id = "test-project"
        job_id = "test-job-id"
        job_path = f"projects/{project_id}/jobs/{job_id}"

        job_cancelled = {}

        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.cancel.return_value.execute.return_value
        ) = job_cancelled

        cancel_job_response = self.hook.cancel_job(job_id=job_id, project_id=project_id)

        assert cancel_job_response == job_cancelled
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().jobs().cancel(name=job_path),
                mock.call().projects().jobs().cancel().execute(num_retries=5),
            ],
            any_order=True,
        )

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_cancel_mlengine_job_nonexistent_job(self, mock_get_conn):
        project_id = "test-project"
        job_id = "test-job-id"
        job_cancelled = {}

        error_job_does_not_exist = HttpError(resp=mock.MagicMock(status=404), content=b"Job does not exist")

        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.cancel.return_value.execute.side_effect
        ) = error_job_does_not_exist
        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.cancel.return_value.execute.return_value
        ) = job_cancelled

        with pytest.raises(HttpError):
            self.hook.cancel_job(job_id=job_id, project_id=project_id)

    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_cancel_mlengine_job_completed_job(self, mock_get_conn):
        project_id = "test-project"
        job_id = "test-job-id"
        job_path = f"projects/{project_id}/jobs/{job_id}"
        job_cancelled = {}

        error_job_already_completed = HttpError(
            resp=mock.MagicMock(status=400), content=b"Job already completed"
        )

        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.cancel.return_value.execute.side_effect
        ) = error_job_already_completed
        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.cancel.return_value.execute.return_value
        ) = job_cancelled

        cancel_job_response = self.hook.cancel_job(job_id=job_id, project_id=project_id)

        assert cancel_job_response == job_cancelled
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().jobs().cancel(name=job_path),
                mock.call().projects().jobs().cancel().execute(num_retries=5),
            ],
            any_order=True,
        )


class TestMLEngineHookWithDefaultProjectId:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.hook = hook.MLEngineHook()

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_create_version(self, mock_get_conn, mock_project_id):
        model_name = "test-model"
        version_name = "test-version"
        version = {"name": version_name}
        operation_path = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/operations/test-operation"
        model_path = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/models/{model_name}"
        operation_done = {"name": operation_path, "done": True}

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.versions.return_value.create.return_value.execute.return_value
        ) = version
        (
            mock_get_conn.return_value.projects.return_value.operations.return_value.get.return_value.execute.return_value
        ) = {"name": operation_path, "done": True}

        create_version_response = self.hook.create_version(
            model_name=model_name, version_spec=version, project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST
        )

        assert create_version_response == operation_done
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().versions().create(body=version, parent=model_path),
                mock.call().projects().models().versions().create().execute(num_retries=5),
                mock.call().projects().operations().get(name=version_name),
                mock.call().projects().operations().get().execute(num_retries=5),
            ],
            any_order=True,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_set_default_version(self, mock_get_conn, mock_project_id):
        model_name = "test-model"
        version_name = "test-version"
        operation_path = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/operations/test-operation"
        version_path = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/models/{model_name}/versions/{version_name}"
        operation_done = {"name": operation_path, "done": True}

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.versions.return_value.setDefault.return_value.execute.return_value
        ) = operation_done

        set_default_version_response = self.hook.set_default_version(
            model_name=model_name,
            version_name=version_name,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )

        assert set_default_version_response == operation_done
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().versions().setDefault(body={}, name=version_path),
                mock.call().projects().models().versions().setDefault().execute(num_retries=5),
            ],
            any_order=True,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.time.sleep")
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_list_versions(self, mock_get_conn, mock_sleep, mock_project_id):
        model_name = "test-model"
        model_path = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/models/{model_name}"
        version_names = [f"ver_{ix}" for ix in range(3)]
        response_bodies = [
            {"nextPageToken": f"TOKEN-{ix}", "versions": [ver]} for ix, ver in enumerate(version_names)
        ]
        response_bodies[-1].pop("nextPageToken")

        pages_requests = [mock.Mock(**{"execute.return_value": body}) for body in response_bodies]
        versions_mock = mock.Mock(
            **{"list.return_value": pages_requests[0], "list_next.side_effect": pages_requests[1:] + [None]}
        )

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.versions.return_value
        ) = versions_mock

        list_versions_response = self.hook.list_versions(
            model_name=model_name, project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST
        )

        assert list_versions_response == version_names
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().versions().list(pageSize=100, parent=model_path),
                mock.call().projects().models().versions().list().execute(num_retries=5),
            ]
            + [
                mock.call()
                .projects()
                .models()
                .versions()
                .list_next(previous_request=pages_requests[i], previous_response=response_bodies[i])
                for i in range(3)
            ],
            any_order=True,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_delete_version(self, mock_get_conn, mock_project_id):
        model_name = "test-model"
        version_name = "test-version"

        operation_path = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/operations/test-operation"
        version_path = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/models/{model_name}/versions/{version_name}"
        version = {"name": operation_path}
        operation_not_done = {"name": operation_path, "done": False}
        operation_done = {"name": operation_path, "done": True}

        (
            mock_get_conn.return_value.projects.return_value.operations.return_value.get.return_value.execute.side_effect
        ) = [operation_not_done, operation_done]

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.versions.return_value.delete.return_value.execute.return_value
        ) = version

        delete_version_response = self.hook.delete_version(
            model_name=model_name, version_name=version_name, project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST
        )

        assert delete_version_response == operation_done
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().versions().delete(name=version_path),
                mock.call().projects().models().versions().delete().execute(num_retries=5),
                mock.call().projects().operations().get(name=operation_path),
                mock.call().projects().operations().get().execute(num_retries=5),
            ],
            any_order=True,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_create_model(self, mock_get_conn, mock_project_id):
        model_name = "test-model"
        model = {
            "name": model_name,
        }
        project_path = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}"

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.create.return_value.execute.return_value
        ) = model

        create_model_response = self.hook.create_model(model=model, project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST)

        assert create_model_response == model
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().create(body=model, parent=project_path),
                mock.call().projects().models().create().execute(num_retries=5),
            ]
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_get_model(self, mock_get_conn, mock_project_id):
        model_name = "test-model"
        model = {"model": model_name}
        model_path = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/models/{model_name}"

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.get.return_value.execute.return_value
        ) = model

        get_model_response = self.hook.get_model(
            model_name=model_name, project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST
        )

        assert get_model_response == model
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().get(name=model_path),
                mock.call().projects().models().get().execute(num_retries=5),
            ]
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_delete_model(self, mock_get_conn, mock_project_id):
        model_name = "test-model"
        model = {"model": model_name}
        model_path = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/models/{model_name}"

        (
            mock_get_conn.return_value.projects.return_value.models.return_value.delete.return_value.execute.return_value
        ) = model

        self.hook.delete_model(model_name=model_name, project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST)

        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().models().delete(name=model_path),
                mock.call().projects().models().delete().execute(num_retries=5),
            ]
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.time.sleep")
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_create_mlengine_job(self, mock_get_conn, mock_sleep, mock_project_id):
        job_id = "test-job-id"
        project_path = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}"
        job_path = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/jobs/{job_id}"
        new_job = {
            "jobId": job_id,
            "foo": 4815162342,
        }
        job_succeeded = {
            "jobId": job_id,
            "state": "SUCCEEDED",
        }
        job_queued = {
            "jobId": job_id,
            "state": "QUEUED",
        }

        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.create.return_value.execute.return_value
        ) = job_queued
        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.get.return_value.execute.side_effect
        ) = [job_queued, job_succeeded]

        create_job_response = self.hook.create_job(job=new_job, project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST)

        assert create_job_response == job_succeeded
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().jobs().create(body=new_job, parent=project_path),
                mock.call().projects().jobs().get(name=job_path),
                mock.call().projects().jobs().get().execute(num_retries=5),
            ],
            any_order=True,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineHook.get_conn")
    def test_cancel_mlengine_job(self, mock_get_conn, mock_project_id):
        job_id = "test-job-id"
        job_path = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/jobs/{job_id}"

        job_cancelled = {}

        (
            mock_get_conn.return_value.projects.return_value.jobs.return_value.cancel.return_value.execute.return_value
        ) = job_cancelled

        cancel_job_response = self.hook.cancel_job(job_id=job_id, project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST)

        assert cancel_job_response == job_cancelled
        mock_get_conn.assert_has_calls(
            [
                mock.call().projects().jobs().cancel(name=job_path),
                mock.call().projects().jobs().cancel().execute(num_retries=5),
            ],
            any_order=True,
        )


def session():
    return mock.Mock()


@pytest.mark.asyncio
@mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineAsyncHook._get_link")
async def test_async_get_job_should_execute_successfully(mocked_link):
    await mlengine_hook.get_job(project_id=PROJECT_ID, job_id=JOB_ID, session=session)
    mocked_link.assert_awaited_once_with(
        url=f"https://ml.googleapis.com/v1/projects/{PROJECT_ID}/jobs/{JOB_ID}", session=session
    )


@pytest.mark.asyncio
async def test_async_get_job_should_fail_if_no_job_id():
    with pytest.raises(
        AirflowException, match=r"An unique job id is required for Google MLEngine training job."
    ):
        await mlengine_hook.get_job(project_id=PROJECT_ID, job_id=None, session=session)


@pytest.mark.asyncio
async def test_async_get_job_should_fail_if_no_project_id():
    with pytest.raises(AirflowException, match=r"Google Cloud project id is required."):
        await mlengine_hook.get_job(project_id=None, job_id=JOB_ID, session=session)


@pytest.mark.asyncio
@mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineAsyncHook.get_job")
async def test_async_get_job_status_should_execute_successfully(mocked_get):
    mocked_get.return_value = ClientResponse(
        "get",
        URL(f"https://ml.googleapis.com/v1/projects/{PROJECT_ID}/jobs/{JOB_ID}"),
        request_info=mock.Mock(),
        writer=mock.Mock(),
        continue100=None,
        timer=TimerNoop(),
        traces=[],
        loop=mock.Mock(),
        session=None,
    )
    mocked_get.return_value._headers = {"Content-Type": "application/json;charset=cp1251"}
    mocked_get.return_value._body = b'{"state": "SUCCEEDED"}'

    job_status = await mlengine_hook.get_job_status(job_id=JOB_ID, project_id=PROJECT_ID)
    mocked_get.assert_awaited_once()
    assert job_status == "success"


@pytest.mark.asyncio
@mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineAsyncHook.get_job")
async def test_async_get_job_status_still_running_should_execute_successfully(mocked_get):
    """Assets that the MLEngineAsyncHook returns a pending response when job is still in running state"""
    mocked_get.return_value = ClientResponse(
        "get",
        URL(f"https://ml.googleapis.com/v1/projects/{PROJECT_ID}/jobs/{JOB_ID}"),
        request_info=mock.Mock(),
        writer=mock.Mock(),
        continue100=None,
        timer=TimerNoop(),
        traces=[],
        loop=mock.Mock(),
        session=None,
    )
    mocked_get.return_value._headers = {"Content-Type": "application/json;charset=cp1251"}
    mocked_get.return_value._body = b'{"state": "RUNNING"}'

    job_status = await mlengine_hook.get_job_status(job_id=JOB_ID, project_id=PROJECT_ID)
    mocked_get.assert_awaited_once()
    assert job_status == "pending"


@pytest.mark.asyncio
@mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineAsyncHook.get_job")
async def test_async_get_job_status_with_oserror_should_execute_successfully(mocked_get):
    """Assets that the MLEngineAsyncHook returns a pending response when OSError is raised"""
    mocked_get.side_effect = OSError()

    job_status = await mlengine_hook.get_job_status(job_id=JOB_ID, project_id=PROJECT_ID)
    mocked_get.assert_awaited_once()
    assert job_status == "pending"


@pytest.mark.asyncio
@mock.patch("airflow.providers.google.cloud.hooks.mlengine.MLEngineAsyncHook.get_job")
async def test_async_get_job_status_with_exception_should_execute_successfully(mocked_get, caplog):
    """Assets that the logging is done correctly when MLEngineAsyncHook raises Exception"""
    caplog.set_level(logging.INFO)
    mocked_get.side_effect = Exception()

    await mlengine_hook.get_job_status(job_id=JOB_ID, project_id=PROJECT_ID)
    assert "Query execution finished with errors..." in caplog.text


@pytest.mark.asyncio
async def test_async_get_job_status_should_fail_if_no_job_id():
    with pytest.raises(
        AirflowException, match=r"An unique job id is required for Google MLEngine training job."
    ):
        await mlengine_hook.get_job_status(project_id=PROJECT_ID, job_id=None)


@pytest.mark.asyncio
async def test_async_get_job_status_should_fail_if_no_project_id():
    with pytest.raises(AirflowException, match=r"Google Cloud project id is required."):
        await mlengine_hook.get_job_status(project_id=None, job_id=JOB_ID)
